/*******************************************************************************
**	generic fixed clock chip driver
********************************************************************************
**
   Copyright (C) 1995-1997 by Steffen Seeger
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
********************************************************************************
**
**	$Log: fixed.inc,v $
**	Revision 1.7  1998/12/03 22:56:19  becka
**	Fixed a few uglinesses, that clash with suidkgi.
**	
**	Revision 1.6  1998/11/19 02:43:04  marcus
**	License change
**	
**	Revision 1.5  1998/11/10 00:13:06  becka
**	Suidkgi commited. Compiles.
**	
**	Revision 1.4  1998/10/28 05:05:27  brian
**	Did clock portion of cmd --> *cmd check_mode change, and some
**	uint/sint namespace fixing (still much to do on that in clock/)
**	
**	Revision 1.3  1998/10/25 02:05:09  marcus
**	Namespace cleanup to make KGI/LibGGI interaction possible
**	
**	Revision 1.2  1998/08/19 20:47:30  marcus
**	Changed ggi/ *.h includes to kgi/ *.h includes
**	Fixed most of the remaining graphics -> accel changes
**	
**	Revision 1.1  1998/07/07 08:01:52  taylorj
**	*** empty log message ***
**	
**	Revision 1.11  1997/12/14 21:55:53  becka
**	uint/sint namespace fixed ... please test ...
**	
**	Revision 1.10  1997/11/08 15:12:40  becka
**	Namespace cleanup ... probably broke all and everything ...
**	
**	Revision 1.9  1997/10/03 11:26:51  seeger
**	major:	overall cleanup to GGI coding style.
**	
*/

#include <kgi/maintainers.h>
#undef	MAINTAINER
#define	MAINTAINER	Steffen_Seeger
#define	COMMON_REV	"$Revision: 1.7 $"

#if !(CLOCKVERSION == 01) 

	#warning	kgi_clock data structures changed, check this code !!
#endif

#define	FREQ	kgi_fixed_clock.clock.freq
#define	INVALIDCLK	-1

/*
**	KGI clock driver interface
*/

int kgim_clock_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev " DRIVER_REV ",  common code:" COMMON_REV);
	
	card->clock = init_clock();

	return card->clock ? EOK : -E(CLOCK, UNKNOWN);
}

void kgim_clock_done(struct kgi_display *dpy)
{
	done_clock();
	NOTICE(DRIVER_NAME " unloaded");
}

int kgim_clock_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
	 enum kgi_tm_cmd *cmd)
{
	kgi_sint i;
	kgi_sint newfreq = (*cmd == CMD_PROPOSE) 
			? 0x7FFFFFFF
			: mode->dclk * mode->clk.mul / mode->clk.div;
	kgi_sint best = 0;
	kgi_sint delta = FREQ[best] - newfreq;
	
	for (i = 1; (i < MAX_FIXED_CLOCKS) && FREQ[i]; i++)  {

		if (abs(FREQ[i] - newfreq) < abs(delta)) {

			best  = i;
			delta = FREQ[best] - newfreq;
		}
	}

	DEBUG1("best == %i, %i Hz, delta = %i\n", best, FREQ[best], delta);

	switch (*cmd) {

	case CMD_PROPOSE:
		mode->clock = INVALIDCLK;
		return EOK;

	case CMD_LOWER:

		if ((delta > 0) && (abs(delta) >= (2*newfreq/100))) {

			best++;
			if ((best == MAX_FIXED_CLOCKS) || !FREQ[best]) {

				return -E(CLOCK, UNKNOWN);
			}
		}

		mode->dclk = FREQ[best] * mode->clk.div / mode->clk.mul;
		DEBUG1("request for %i Hz, did %i Hz", newfreq, FREQ[best]);
		return EOK;

	case CMD_RAISE:

		if ((delta < 0) && (abs(delta) >= (2*newfreq/100))) {

			best--;
			if (best < 0) {

				return -E(CLOCK, UNKNOWN);
			}
		}

		mode->dclk = FREQ[best] * mode->clk.div / mode->clk.mul;
	
		DEBUG1("request for %i Hz, did %i Hz", newfreq, FREQ[best]);
		return EOK;

	case CMD_CHECK:

		if ((abs(delta) < (2*newfreq/100)) && FREQ[best]) {

			mode->clock = best;
			return EOK;

		} else {

			mode->clock = INVALIDCLK; 
			DEBUG2("invalid clock request");
			return -E(CLOCK, UNKNOWN);
		}

	default:			/* this should not happen ... */
		INTERNAL_ERROR;
		return -E(CLOCK, UNKNOWN);
	}
}

void kgim_clock_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{
	if (mode->clock == INVALIDCLK) {

		INTERNAL_ERROR;
	}

	CLK_out8(dpy, clkindex[mode->clock],0);
}

int kgim_clock_command(struct kgi_graphic *graph,kgi_uint cmd,kgi_uint iobuff)
{
	return -EINVAL;
}
